/*
 * Copyright (c) 1992 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: psfile.cc,v 1.10 94/12/04 07:33:41 van Exp $ (LBL)";
#endif

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include <osfcn.h>
#include <string.h>
#include <ctype.h>

#include "psfile.h"

#ifndef MIN
#define MIN(a, b) ((a) < (b) ? (a) : (b))
#endif

PSfile::PSfile(int f)
{
	fd = dup(f);

	npage = -1;
	info = 0;
	if (dostat() < 0)
		return;
	pageinfo pi;
	npage = scanhdr(pi);
	if (npage < 0)
		return;
	/*
	 * Allocate enough room for each page plus the preamble.
	 * It is convenient to number the premable 0, since pages
	 * are numbered starting from 1.
	 */
	info = new pageinfo[npage + 1];
	info[0] = pi;
	for (int i = 1; i <= npage; ++i) {
		info[i].length = -1;
		info[i].offset = -1;
	}
}

PSfile::~PSfile()
{
	if (info)
		delete info;
	(void)close(fd);
}

int PSfile::dostat()
{
	struct stat st;
	if (fstat(fd, &st) < 0)
		return (-1);
	filesize = st.st_size;
	return (1);
}

/*
 * Return true if null-terminated string, n, matches string s,
 * which is not necessarily null-terminated.
 * min(len, strlen(n)) bytes are examined.
 */
int PSfile::match(const char* nt, const char* s, int len) const
{
	while (--len >= 0) {
		if (*nt == 0)
			return (1);
		if (*nt++ != *s++)
			return (0);
	}
	return (1);
}

int PSfile::fsearch(const char* cp, int len, const char* target) const
{
	const char* base = cp;
	while (1) {
		if (match(target, cp, len))
			return (cp - base);
		do
			if (--len <= 0)
				return (-1);
		while (*cp++ != '\n');
	}
}

int PSfile::bsearch(const char* cp, int len, const char* target) const
{
	const char* base = cp;
	cp += len;
	while (cp > base) {
		len = 0;
		while (cp > base) {
			if (*--cp == '\n') {
				++cp;
				break;
			}
			++len;
		}
		if (len != 0 && match(target, cp, len))
			return (cp - base);
		--cp;
	}
	return (-1);
}

off_t PSfile::nextpage(off_t offset)
{
	char file[4096];
	while (1) {
		int len = MIN(filesize - offset, int(sizeof(file)));
		lseek(fd, offset, SEEK_SET);
		if (read(fd, file, len) != len)
			return (-1);
		int k = fsearch(file, len, "%%Page:");
		if (k > 0) {
			int pnum = atoi(file + k + 8);
			if (pnum >= lastnum) {
				lastnum = pnum;
				return (offset + k);
			}
		}
		if (offset + len >= filesize)
			return (-1);
		offset +=  len - 8;
	}
}

/*
 * Check that file is DSC conformant.  Determine number of pages and
 * length of preamble by finding corresponding comment strings.
 * -1 is returned for bogus files.
 */
int PSfile::scanhdr(pageinfo& pi)
{
	int n, off;
	off_t poff;
	char* num;
	char file[4096];

	int len = MIN(filesize, int(sizeof(file)));
	if (read(fd, file, len) != len)
		return (-1);
	if (strncmp(file, "%!PS-Adobe-", 11) != 0) {
		/*
		 * the file isn't conforming -- if it's postscript,
		 * say it has one page.
		 */
		if (strncmp(file, "%!", 2) != 0)
			return (-1);

		goto one_page;
	}
	off = fsearch(file, len, "%%Pages:");
	if (off < 0)
		goto one_page;
	for (off += 8; file[off] == ' '; ++off) {
	}
	num = file + off;
	if (match("(atend)", num, 7)) {
		if (filesize > int(sizeof(file))) {
			off_t noff = filesize - len;
			lseek(fd, noff, SEEK_SET);
			if (read(fd, file, len) != len)
				goto one_page;
		}
		n = bsearch(file, len, "%%Pages:");
		if (n < 0)
			goto one_page;
		for (n += 8; file[n] == ' '; ++n) {
		}
		num = file + n;
	}
	n = atoi(num);
	if (n <= 0)
		goto one_page;
	/*
	 * Look for start of first page.
	 * This gives us the size of the preamble.
	 */
	lastnum = 0;
	poff = nextpage(0);
	if (poff <= off)
		goto one_page;
	pi.offset = 0;
	pi.length = poff;
	return (n);
one_page:
	pi.offset = 0;
	pi.length = 2; /*XXX*/
	return (1);
}

void PSfile::scaninfo(int pageno)
{
	/* check for special 'non-conforming ps' marker */
	if (info[0].length == 2 && pageno == 1) {
		info[1].offset = info[0].length;
		info[1].length = filesize - info[0].length;
		return;
	}
	int i = pageno;
	while (info[i].length < 0)
		--i;
	off_t offset = info[i].offset + info[i].length;
	while (++i <= pageno) {
		info[i].offset = offset;
		if (i == npage) {
			info[i].length = filesize - offset;
			break;
		}
		off_t noff = nextpage(offset + 8);
		if (noff <= 0) {
			/* XXX bogus document */
			do {
				info[i].length = 0;
			} while (++i <= npage);
			return;
		}
		info[i].length = noff - offset;
		offset = noff;
	}
}

int PSfile::page_length(int pageno)
{
	register int l = info[pageno].length;
	if (l < 0)
		scaninfo(pageno);
	l = info[pageno].length;
	if (l > 0)
		l += info[0].length;
	return (l);
}

static char* pbuf;
int pbufsize;

const char* PSfile::page(int pageno, int& len)
{
	off_t o;
	int l;
	pageinfo& pi = info[pageno];
	if (pi.length < 0)
		scaninfo(pageno);
	o = pi.offset;
	l = pi.length;
	if (pageno == 1) {
		/* adjust page 1 so it includes the preamble */
		l += pi.offset;
		o = 0;
	}
	if (l > pbufsize) {
		if (pbuf == 0)
			pbuf = (char*)malloc(l);
		else
			pbuf = (char*)realloc(pbuf, l);
		pbufsize = l;
	}
	lseek(fd, o, SEEK_SET);
	read(fd, pbuf, l);
	len = l;
	return (pbuf);
}
