/*
 * Copyright (c) 1992 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: pssplit.cc,v 1.3 93/10/17 19:56:45 van Locked $ (LBL)";
#endif

#include <osfcn.h>
#include <sys/file.h>
#include "psfile.h"

void usage()
{
	fprintf(stderr, "-usage: pssplit [-lz] [-p fileprefix] [-v] [file]\n");
	exit(1);
}

char* fname = "<stdin>";
char* fprefix = "p";
int verbose = 0;
int lz = 0;

void pswrite(FILE*, const char*, int);
void lzinit(FILE* f);
void lzdone(FILE* f);
void lzwrite(FILE* f, const char* cp, int len);

/*
 * skip the leading comments so we don't lie about # pages, bounding
 * box, fonts, etc.
 */
const char* skip_comments(const char* cp, int& len)
{
	/* 0 = saw eol, 1 = skipping to eol */
	register int state = 0;
	for (const char* ep = cp + len; cp < ep; ++cp) {
		switch (*cp) {
		case '%':
			state = 1;
			break;
		case '\n':
		case '\r':
			state = 0;
			break;
		default:
			if (state == 0) {
				len = ep - cp;
				return (cp);
			}
			break;
		}
	}
	len = 0;
	return (cp);
}

main(int argc, char** argv)
{
	int fd;

	if (argc < 2)
		usage();

	for (int i = 1; i < argc; ++i) {
		if (argv[i][0] != '-' || argv[i][1] == 0)
			break;
		if (strcmp(argv[i], "-lz") == 0)
			lz = 1;
		else if (strcmp(argv[i], "-v") == 0)
			verbose = 1;
		else if (strcmp(argv[i], "-p") == 0) {
			if (++i >= argc)
				usage();
			fprefix = argv[i];
		} else
			usage();
	}
	if (i < argc && argv[i][0] != '-') {
		fname = argv[i];
		fd = open(fname, O_RDONLY);
		if (fd < 0) {
			perror(fname);
			exit(1);
		}
	} else
		fd = fileno(stdin);

	PSfile* ps = new PSfile(fd);
	if (!ps->valid()) {
		fprintf(stderr, "-%s: %s is not a valid postscript file\n",
			argv[0], fname);
		exit(1);
	}
	int npage = ps->npages();
	if (npage == 1) {
		fprintf(stderr, "-%s: %s has only one page\n", argv[0], fname);
		exit(2);
	}
	int plen = ps->preamble_length();
	if (verbose)
		fprintf(stderr, "%s: %d pages, %d byte preamble\n", fname,
			npage, plen);

	int foundone = 0;
	for (i = 1; i <= npage; ++i) {
		char pname[1024];
		struct stat s;
		sprintf(pname, "%s%02d.ps", fprefix, i);
		if (stat(pname, &s) != -1 || errno != ENOENT) {
			fprintf(stderr,
				"-%s: file %s exists - won't overwrite\n",
				argv[0], pname);
			foundone = 1;
		}
	}
	if (foundone)
		exit(4);

	for (i = 1; i <= npage; ++i) {
		char pname[1024];
		sprintf(pname, "%s%02d.ps", fprefix, i);
		FILE* wf = fopen(pname, "w");
		if (wf == NULL) {
			perror(pname);
			exit(3);
		}
		fprintf(wf, "%%!PS-\n");
		if (lz)
			lzinit(wf);
		const char* cp;
		int len;
		if (i > 1) {
			cp = ps->page(0, len);
			cp = skip_comments(cp, len);
			pswrite(wf, cp, len);
			cp = ps->page(i, len);
			pswrite(wf, cp, len);
		} else {
			cp = ps->page(i, len);
			cp = skip_comments(cp, len);
			pswrite(wf, cp, len);
		}
		if (lz)
			lzdone(wf);

		if (verbose)
			fprintf(stderr, "  page %2d -> \"%s\" (%d bytes)\n",
				i, pname, int(ftell(wf)));
		fclose(wf);
	}
	delete ps;
	exit(0);
}

void pswrite(FILE* f, const char* cp, int len)
{
	if (lz)
		lzwrite(f, cp, len);
	else
		fwrite(cp, 1, len, f);
}

/*
 * Various constants
 */
#define EOD		257	/* End of Data marker */
#define CLR		256	/* Clear Table Marker */
#define MAXSTRINGS	4096	/* using 12 bit max codes */

/*
 * This is the structure for a single entry into the string table
 */
typedef struct {
	short prefix;
	short suffix;
	short next;
} stringrec;

int codeWidth = 9;
int bitsInAcc = 0;		/* Used by code send function */
unsigned long acc;		/* =bits left over from a sendcode */
int nextIndex;			/* Used by newentry function */
stringrec stringarray[4096];	/* This is the string table */
short lastCode;

void Initialize(void);
void SendCode(short theCode, FILE* f);
int FindString(short *prefix, short suffix);
void AddString(short prefix, short suffix, FILE* f);

void 
Initialize()
{
	int i;

	for (i = 0; i < 256; i++) {
		stringarray[i].prefix = -1;
		stringarray[i].suffix = (char)i;
		stringarray[i].next = -1;
	}
	nextIndex = EOD + 1;
	codeWidth = 9;
}

void 
SendCode(short theCode, FILE* f)
{
	acc += (long) theCode << (32 - codeWidth - bitsInAcc);
	bitsInAcc += codeWidth;
	while (bitsInAcc >= 8) {
		fputc(acc >> 24, f);
		acc = acc << 8;
		bitsInAcc -= 8;
	}
	if (theCode == CLR) {
		Initialize();
	}
	if (theCode == EOD && bitsInAcc != 0) {
		fputc(acc >> 24, f);
	}
}

void lzinit(FILE* f) {
	lastCode = -1;
	bitsInAcc = 0;
	acc = 0;
	fprintf(f, "currentfile /LZWDecode filter cvx exec\n");
	Initialize();
	SendCode(CLR, f);
}

void lzdone(FILE* f) {
	SendCode(lastCode, f);
	SendCode(EOD, f);
}

void lzwrite(FILE* f, const char* cp, int len)
{
	short   thisCode;

	if (lastCode == -1) {
		lastCode = *cp++;
		--len;
	}
	while (--len >= 0) {
		thisCode = *cp++;
		if (FindString(&lastCode, thisCode) == 0) {
			SendCode(lastCode, f);
			AddString(lastCode, thisCode, f);
			lastCode = thisCode;
		}
	}
}

int 
FindString(short *prefix, short suffix)
{
	short   i = *prefix;

	while (i != -1) {
		if (stringarray[i].prefix == *prefix &&
		    stringarray[i].suffix == suffix) {
			*prefix = i;
			return (1);
		} else {
			i = stringarray[i].next;
		}
	}
	return (0);
}

void 
AddString(short prefix, short suffix, FILE* f)
{

	stringarray[nextIndex].prefix = prefix;
	stringarray[nextIndex].suffix = suffix;
	stringarray[nextIndex].next = stringarray[prefix].next;
	stringarray[prefix].next = nextIndex;

	if (++nextIndex >> codeWidth) {
		if (++codeWidth > 12) {
			--codeWidth;
			SendCode(CLR, f);
		}
	}
}
